use std::{
    cell::RefCell,
    collections::{BTreeMap, HashSet},
    rc::Rc,
};

use itertools::{EitherOrBoth, Itertools};
use rand::{seq::IndexedRandom, RngCore};
use serde::{Deserialize, Serialize};

use crate::choice_function::ChoiceFunction;

#[derive(Clone, Deserialize, Serialize, Debug)]
pub struct Genotype {
    pub chromosomes: BTreeMap<String, Chromosomes>,
}
type Chromosomes = (String, String);

fn merge(rng: &mut impl RngCore, one: &str, two: &str) -> String {
    let mut results = String::new();
    for pair in one.chars().zip_longest(two.chars()) {
        results.push(match pair {
            EitherOrBoth::Both(l, r) => *[l, r].choose(rng).unwrap(),
            EitherOrBoth::Left(l) => l,
            EitherOrBoth::Right(r) => r,
        })
    }
    return results;
}

impl Genotype {
    pub fn breed(&self, other: &Genotype) -> Genotype {
        let mut rng = rand::rng();
        let mut chromosomes: HashSet<String> = HashSet::new();
        chromosomes.extend(self.chromosomes.keys().cloned());
        chromosomes.extend(other.chromosomes.keys().cloned());

        let mut result: BTreeMap<String, Chromosomes> = BTreeMap::new();
        let placeholder = (String::new(), String::new());
        for k in chromosomes {
            let mine = self.chromosomes.get(&k).unwrap_or(&placeholder);
            let theirs = self.chromosomes.get(&k).unwrap_or(&placeholder);

            let mine_merged = merge(&mut rng, &mine.0, &mine.1);
            let theirs_merged = merge(&mut rng, &theirs.0, &theirs.1);
            result.insert(k, (mine_merged, theirs_merged));
        }
        return Genotype {
            chromosomes: result,
        };
    }

    pub fn generate<T: Genetic>(choice_function: ChoiceFunction) -> Self {
        let root = RootTranscriber::Generator(Generator {
            choice_function,
            progress: BTreeMap::new(),
        });
        let root_cell = Rc::new(RefCell::new(root));
        let transcriber = Transcriber {
            root: root_cell.clone(),
            chromosome: String::from("base"),
        };
        let mut base = T::default();
        base.transcribe(transcriber);
        match &*root_cell.borrow() {
            RootTranscriber::Generator(generator) => {
                let chromosomes = generator.progress.clone();
                return Genotype { chromosomes };
            }
            _ => panic!("a generator should not able to turn into a loader"),
        };
    }

    pub fn load<T: Genetic>(&self) -> T {
        let mut progress = BTreeMap::new();
        for (key, (lhs, rhs)) in self.chromosomes.iter() {
            progress.insert(
                key.clone(),
                (0, Vec::from_iter(lhs.chars()), Vec::from_iter(rhs.chars())),
            );
        }
        let root = RootTranscriber::Loader(Loader { progress });
        let root_cell = Rc::new(RefCell::new(root));
        let transcriber = Transcriber {
            root: root_cell.clone(),
            chromosome: String::from("base"),
        };
        let mut base = T::default();
        base.transcribe(transcriber);
        return base;
    }
}

enum RootTranscriber {
    Generator(Generator),
    Loader(Loader),
}

impl RootTranscriber {
    pub fn express<T: Clone>(
        &mut self,
        chromosome: &str,
        field: &mut T,
        possibilities: &[(char, T)],
    ) {
        match self {
            RootTranscriber::Generator(generator) => {
                generator.express(chromosome, field, possibilities)
            }
            RootTranscriber::Loader(loader) => loader.express(chromosome, field, possibilities),
        }
    }
}

#[derive(Clone)]
pub struct Transcriber {
    root: Rc<RefCell<RootTranscriber>>,
    chromosome: String,
}

impl Transcriber {
    pub fn express<T: Clone>(&mut self, field: &mut T, possibilities: &[(char, T)]) {
        self.root
            .borrow_mut()
            .express(&self.chromosome, field, possibilities)
    }

    pub fn nest(&self, suffix: &str) -> Transcriber {
        return Transcriber {
            root: self.root.clone(),
            chromosome: format!("{}.{}", self.chromosome, suffix),
        };
    }
}

pub trait Genetic: Default {
    fn transcribe(&mut self, t: Transcriber);
}

struct Generator {
    choice_function: ChoiceFunction,
    progress: BTreeMap<String, Chromosomes>,
}

impl Generator {
    pub fn express<T>(&mut self, chromosome: &str, _field: &mut T, possibilities: &[(char, T)]) {
        let lhs = self.choice_function.choose(possibilities).0;
        let rhs = self.choice_function.choose(possibilities).0;

        let progress = self
            .progress
            .entry(chromosome.to_string())
            .or_insert((String::new(), String::new()));

        progress.0.push(lhs);
        progress.1.push(rhs);
    }
}

struct Loader {
    progress: BTreeMap<String, (usize, Vec<char>, Vec<char>)>,
}

impl Loader {
    pub fn express<T: Clone>(
        &mut self,
        chromosome: &str,
        field: &mut T,
        possibilities: &[(char, T)],
    ) {
        let (mut lhs, mut rhs) = ('?', '?');
        if let Some((ix, lhs_buf, rhs_buf)) = self.progress.get_mut(chromosome) {
            lhs = lhs_buf.get(*ix).cloned().unwrap_or('?');
            rhs = rhs_buf.get(*ix).cloned().unwrap_or('?');
            *ix += 1;
        }

        let mut expressed = lhs; // lhs wins if not dominated
        if dominance(rhs) > dominance(lhs) {
            expressed = rhs;
        }
        for (allele, value) in possibilities.iter() {
            if *allele == expressed {
                *field = value.clone();
                return;
            }
        }
    }
}

fn dominance(allele: char) -> u8 {
    if allele.is_ascii_lowercase() {
        return 1;
    }
    if allele.is_ascii_uppercase() {
        return 2;
    }
    return 0;
}
